[llvm] r270892 - [MemCpyOpt] Don't perform callslot optimization across may-throw calls

David Majnemer via llvm-commits llvm-commits at lists.llvm.org
Thu May 26 12:24:24 PDT 2016


Author: majnemer
Date: Thu May 26 14:24:24 2016
New Revision: 270892

URL: http://llvm.org/viewvc/llvm-project?rev=270892&view=rev
Log:
[MemCpyOpt] Don't perform callslot optimization across may-throw calls

An exception could prevent a store from occurring but MemCpyOpt's
callslot optimization would fire anyway, causing the store to occur.

This fixes PR27849.

Added:
    llvm/trunk/test/Transforms/MemCpyOpt/callslot_throw.ll
Modified:
    llvm/trunk/lib/Transforms/Scalar/MemCpyOptimizer.cpp
    llvm/trunk/test/Transforms/MemCpyOpt/loadstore-sret.ll

Modified: llvm/trunk/lib/Transforms/Scalar/MemCpyOptimizer.cpp
URL: http://llvm.org/viewvc/llvm-project/llvm/trunk/lib/Transforms/Scalar/MemCpyOptimizer.cpp?rev=270892&r1=270891&r2=270892&view=diff
==============================================================================
--- llvm/trunk/test/Transforms/MemCpyOpt/callslot_throw.ll (added)
+++ llvm/trunk/test/Transforms/MemCpyOpt/callslot_throw.ll Thu May 26 14:24:24 2016
@@ -0,0 +1,34 @@
+; RUN: opt -S -memcpyopt < %s | FileCheck %s
+declare void @may_throw(i32* nocapture %x)
+
+; CHECK-LABEL: define void @test1(
+define void @test1(i32* nocapture noalias dereferenceable(4) %x) {
+entry:
+  %t = alloca i32, align 4
+  call void @may_throw(i32* nonnull %t)
+  %load = load i32, i32* %t, align 4
+  store i32 %load, i32* %x, align 4
+; CHECK:       %[[t:.*]] = alloca i32, align 4
+; CHECK-NEXT:  call void @may_throw(i32* {{.*}} %[[t]])
+; CHECK-NEXT:  %[[load:.*]] = load i32, i32* %[[t]], align 4
+; CHECK-NEXT:  store i32 %[[load]], i32* %x, align 4
+  ret void
+}
+
+declare void @always_throws()
+
+; CHECK-LABEL: define void @test2(
+define void @test2(i32* nocapture noalias dereferenceable(4) %x) {
+entry:
+  %t = alloca i32, align 4
+  call void @may_throw(i32* nonnull %t) nounwind
+  %load = load i32, i32* %t, align 4
+  call void @always_throws()
+  store i32 %load, i32* %x, align 4
+; CHECK:       %[[t:.*]] = alloca i32, align 4
+; CHECK-NEXT:  call void @may_throw(i32* {{.*}} %[[t]])
+; CHECK-NEXT:  %[[load:.*]] = load i32, i32* %[[t]], align 4
+; CHECK-NEXT:  call void @always_throws()
+; CHECK-NEXT:  store i32 %[[load]], i32* %x, align 4
+  ret void
+}

Modified: llvm/trunk/test/Transforms/MemCpyOpt/loadstore-sret.ll
URL: http://llvm.org/viewvc/llvm-project/llvm/trunk/test/Transforms/MemCpyOpt/loadstore-sret.ll?rev=270892&r1=270891&r2=270892&view=diff
==============================================================================
--- llvm/trunk/test/Transforms/MemCpyOpt/loadstore-sret.ll (original)
+++ llvm/trunk/test/Transforms/MemCpyOpt/loadstore-sret.ll Thu May 26 14:24:24 2016
@@ -22,4 +22,4 @@ _ZNSt8auto_ptrIiED1Ev.exit:
   ret void
 }
 
-declare void @_Z3barv(%"class.std::auto_ptr"* nocapture sret)
+declare void @_Z3barv(%"class.std::auto_ptr"* nocapture sret) nounwind




More information about the llvm-commits mailing list
--- llvm/trunk/lib/Transforms/Scalar/MemCpyOptimizer.cpp (original)
+++ llvm/trunk/lib/Transforms/Scalar/MemCpyOptimizer.cpp Thu May 26 14:24:24 2016
@@ -496,7 +496,7 @@ static unsigned findCommonAlignment(cons
 
 // This method try to lift a store instruction before position P.
 // It will lift the store and its argument + that anything that
 // lay alias with these.
 // The method returns true if it was successful.
 static bool moveUp(AliasAnalysis &AA, StoreInst *SI, Instruction *P) {
   // If the store alias this position, early bail out.
@@ -675,6 +675,8 @@ bool MemCpyOpt::processStore(StoreInst *
       if (C) {
         // Check that nothing touches the dest of the "copy" between
         // the call and the store.
+        Value *CpyDest = SI->getPointerOperand()->stripPointerCasts();
+        bool CpyDestIsLocal = isa<AllocaInst>(CpyDest);
         AliasAnalysis &AA = getAnalysis<AAResultsWrapperPass>().getAAResults();
         MemoryLocation StoreLoc = MemoryLocation::get(SI);
         for (BasicBlock::iterator I = --SI->getIterator(), E = C->getIterator();
@@ -683,6 +685,12 @@ bool MemCpyOpt::processStore(StoreInst *
             C = nullptr;
             break;
           }
+          // The store to dest may never happen if an exception can be thrown
+          // between the load and the store.
+          if (I->mayThrow() && !CpyDestIsLocal) {
+            C = nullptr;
+            break;
+          }
         }
       }
 
@@ -815,6 +823,10 @@ bool MemCpyOpt::performCallSlotOptzn(Ins
     if (destSize < srcSize)
       return false;
   } else if (Argument *A = dyn_cast<Argument>(cpyDest)) {
+    // The store to dest may never happen if the call can throw.
+    if (C->mayThrow())
+      return false;
+
     if (A->getDereferenceableBytes() < srcSize) {
       // If the destination is an sret parameter then only accesses that are
       // outside of the returned struct type can trap.
@@ -496,7 +496,7 @@ static unsigned findCommonAlignment(cons
 // This method try to lift a store instruction before position P.
 // It will lift the store and its argument + that anything that
-// lay alias with these.
+// may alias with these.
 // The method returns true if it was successful.
 static bool moveUp(AliasAnalysis &AA, StoreInst *SI, Instruction *P) {

Added: llvm/trunk/test/Transforms/MemCpyOpt/callslot_throw.ll
URL: http://llvm.org/viewvc/llvm-project/llvm/trunk/test/Transforms/MemCpyOpt/callslot_throw.ll?rev=270892&view=auto
==============================================================================